"""Helper functions for verifying minisign signatures.

This module is meant to be a quick and dirty means of verifying minisign signatures in Python.
There is no readily accessible Python library at this time. The `py-minisign` repository exists
on GitHub to attempt this - https://github.com/x13a/py-minisign - but it does not exist as a
package on PyPI and also does not appear to be actively maintained. There is a `minisign` package
on PyPI - https://pypi.org/project/minisign/ - but it comes from a different repo and has no
functionality at the time of this writing.

Short of forking the `py-minisign` repo to maintain it and publish a package from it on PyPI,
the actual format for signatures and public keys is simple and so is verifying signatures.

Minisign reference: https://jedisct1.github.io/minisign/

Even still, this module is NOT meant to be used as a library or general purpose minisign
signature verification. It is purpose written to specifically verify minisign signatures that
were created by Phylum. As such, it makes a number of assumptions:

* The Minisign Public Key for Phylum, Inc. will not change between releases
* The files to be verified were created by Phylum, Inc.
* The `.minisig` signature includes a trusted comment and will therefore contain a known number of lines
* The source of the `.minisig` signature is a trusted location, controlled by Phylum, Inc. for it's CLI releases
"""
import base64

from cryptography.exceptions import InvalidSignature, UnsupportedAlgorithm
from cryptography.hazmat.primitives.asymmetric import ed25519

# This is the Minisign Public Key for Phylum, Inc. The matching private key was used to sign the software releases
PHYLUM_MINISIGN_PUBKEY = "RWT6G44ykbS8GABiLXrJrYsap7FCY77m/Jyi0fgsr/Fsy3oLwU4l0IDf"

# The format for a minisign public key is:
#
# base64(<signature_algorithm> || <key_id> || <public_key>)
#
# signature_algorithm: `Ed`
# key_id: 8 random bytes
# public_key: Ed25519 public key
PHYLUM_MINISIGN_PUBKEY_SIG_ALGO = base64.b64decode(PHYLUM_MINISIGN_PUBKEY)[:2]
PHYLUM_MINISIGN_PUBKEY_KEY_ID = base64.b64decode(PHYLUM_MINISIGN_PUBKEY)[2:10]
PHYLUM_MINISIGN_PUBKEY_ED25519 = base64.b64decode(PHYLUM_MINISIGN_PUBKEY)[10:]


def verify_minisig(file_path, sig_path):
    """Verify a given file has a valid minisign signature.

    `file_path` is the path to the file data to verify.
    `sig_path` is the path to the `.minisig` file containing the minisign signature information.

    The public key is an assumed constant, the Minisign Public Key for Phylum, Inc.
    """
    try:
        phylum_public_key = ed25519.Ed25519PublicKey.from_public_bytes(PHYLUM_MINISIGN_PUBKEY_ED25519)
    except UnsupportedAlgorithm as err:
        raise RuntimeError("Ed25519 algorithm is not supported by the OpenSSL version `cryptography` is using") from err

    signature_algorithm, key_id, signature, trusted_comment, global_signature = extract_minisig_elements(sig_path)

    if signature_algorithm != b"Ed":
        raise RuntimeError("Only the legacy `Ed` signature algorithm is used by Phylum currently")

    if key_id != PHYLUM_MINISIGN_PUBKEY_KEY_ID:
        raise RuntimeError("The `key_id` from the `.minisig` signature did not match the `key_id` from the public key")

    # Confirm the trusted comment in the sig_path with the `global_signature` there
    try:
        phylum_public_key.verify(global_signature, signature + trusted_comment)
    except InvalidSignature as err:
        raise RuntimeError("The signature could not be verified") from err

    # Confirm the data from file_path with the signature from the .minisig `sig_path`
    with open(file_path, "rb") as f:
        file_data = f.read()
    try:
        phylum_public_key.verify(signature, file_data)
    except InvalidSignature as err:
        raise RuntimeError("The signature could not be verified") from err


def extract_minisig_elements(sig_path):
    """Extract the elements from a given minisig signature file and return them."""
    # The format for a minisign signature is:
    #
    # untrusted comment: <arbitrary text>
    # base64(<signature_algorithm> || <key_id> || <signature>)
    # trusted_comment: <arbitrary text>
    # base64(<global_signature>)
    #
    # where each line above represents a line from the `.minisig` file and the elements are defined as:
    #
    # signature_algorithm: `Ed` (legacy) or `ED` (hashed)
    # key_id: 8 random bytes, matching the public key
    # signature (legacy): ed25519(<file data>)
    # signature (prehashed): ed25519(Blake2b-512(<file data>))
    # global_signature: ed25519(<signature> || <trusted_comment>)
    trusted_comment_prefix = "trusted comment: "
    trusted_comment_prefix_len = len(trusted_comment_prefix)
    ed25519_signature_len = 64

    with open(sig_path, "rb") as f:
        lines = f.read().splitlines()
    if len(lines) not in (4, 5):
        raise RuntimeError("The .minisig file format expects 4 lines, with an optional blank 5th line")

    decoded_sig_line = base64.b64decode(lines[1])
    signature_algorithm = decoded_sig_line[:2]
    key_id = decoded_sig_line[2:10]
    signature = decoded_sig_line[10:]
    if len(signature) != ed25519_signature_len:
        raise RuntimeError(f"The decoded signature was not {ed25519_signature_len} bytes long")

    trusted_comment = lines[2][trusted_comment_prefix_len:]

    global_signature = base64.b64decode(lines[3])
    if len(global_signature) != ed25519_signature_len:
        raise RuntimeError(f"The global signature was not {ed25519_signature_len} bytes long")

    return signature_algorithm, key_id, signature, trusted_comment, global_signature



"""Provide common data structures for the package."""
from dataclasses import dataclass
from enum import IntEnum
from typing import List


@dataclass(order=True, frozen=True)
class PackageDescriptor:
    """Class for keeping track of packages returned by the `phylum parse` subcommand."""

    name: str
    version: str
    type: str


# Type alias
Packages = List[PackageDescriptor]


@dataclass()
class ProjectThresholdInfo:
    """Class for keeping track of project risk threshold information.

    `threshold`: The risk domain threshold value in use.
    `req_src`: The source of the threshold requirement. For example, `phylum-ci option` for command line options
    """

    threshold: float
    req_src: str


@dataclass(order=True, frozen=True)
class RiskDomain:
    """Class for keeping track of a specific risk domain.

    Each risk domain can be referenced in various ways. See "Phylum Risk Domains" documentation for more detail:
    https://docs.phylum.io/docs/phylum-package-score#risk-domains

    * `output_name`: The descriptive name; useful for output expected to be read by humans
    * `project_name`: Key name returned from a Phylum analysis as known by the overall project threshold mapping
    * `package_name`: Key name returned from a Phylum analysis as known by an individual package riskVectors mapping
    """

    output_name: str
    project_name: str
    package_name: str


class ReturnCode(IntEnum):
    """Integer enumeration to track return codes."""

    SUCCESS = 0
    FAILURE = 1
    INCOMPLETE = 5
    FAILURE_INCOMPLETE = 6



"""Provide constants for use throughout the package."""

# Linux platform support in the CLI was changed from `unknown-linux-musl` to `unknown-linux-gnu` starting with
# v3.8.0-rc2, changing the artifact names available to download and install in a non-backwards compatible manner.
MIN_SUPPORTED_CLI_VERSION = "v3.8.0-rc2"

# Keys are lowercase machine hardware names as returned from `uname -m`.
# Values are the mapped rustc architecture.
SUPPORTED_ARCHES = {
    "aarch64": "aarch64",
    "arm64": "aarch64",
    "x86_64": "x86_64",
    "amd64": "x86_64",
}

# Keys are lowercase operating system name as returned from `uname -s`.
# Values are the mapped rustc platform, which is the vendor-os_type[-environment_type].
SUPPORTED_PLATFORMS = {
    "linux": "unknown-linux-gnu",
    "darwin": "apple-darwin",
}

# Environment variable name to hold the Phylum CLI token used to access the backend API.
# The API token can also be set via the environment variable `PHYLUM_API_KEY`, which will take precedence over
# the `offline_access` parameter in the `settings.yaml` file.
TOKEN_ENVVAR_NAME = "PHYLUM_API_KEY"  # nosec ; this is NOT a hard-coded password

# These are the currently supported lockfiles.
# Keys are the standard lockfile filename, optionally specified with glob syntax.
# Values are the name of the tool that generates the lockfile.
SUPPORTED_LOCKFILES = {
    # Javascript/Typescript
    "package-lock.json": "npm",
    "yarn.lock": "yarn",
    # Ruby
    "Gemfile.lock": "gem",
    # Python
    "requirements.txt": "pip",
    "Pipfile.lock": "pipenv",
    "Pipfile": "pipenv",
    "poetry.lock": "poetry",
    # C#
    "*.csproj": "nuget",
    # Java
    "pom.xml": "mvn",
    "gradle.lockfile": "gradle",
}

# Timeout value, in seconds, to tell the Python Requests package to stop waiting for a response.
# Reference: https://2.python-requests.org/en/master/user/quickstart/#timeouts
REQ_TIMEOUT: float = 10.0



"""Test the minisign signature verification module."""
from phylum.init import sig


def test_phylum_minisign_pubkey():
    """Ensure the minisign public key in use by Phylum has not changed."""
    expected_key = "RWT6G44ykbS8GABiLXrJrYsap7FCY77m/Jyi0fgsr/Fsy3oLwU4l0IDf"
    assert sig.PHYLUM_MINISIGN_PUBKEY == expected_key, "The key should not be changing"


def test_phylum_pubkey_sig_algo():
    """Ensure the Phylum minisign public key signature algorithm is `Ed` (legacy)."""
    assert isinstance(sig.PHYLUM_MINISIGN_PUBKEY_SIG_ALGO, bytes)
    assert sig.PHYLUM_MINISIGN_PUBKEY_SIG_ALGO == b"Ed", "Only the legacy `Ed` signature is used by Phylum currently"


def test_phylum_pubkey_key_id():
    """Ensure the Phylum minisign public key `key_id` has not changed."""
    expected_key_id = b"\xfa\x1b\x8e2\x91\xb4\xbc\x18"
    assert isinstance(sig.PHYLUM_MINISIGN_PUBKEY_KEY_ID, bytes)
    assert sig.PHYLUM_MINISIGN_PUBKEY_KEY_ID == expected_key_id, "The key ID should not be changing"


def test_phylum_ed25519_pubkey():
    """Ensure the Phylum minisign Ed25519 public key has not changed."""
    expected_key = b"\x00b-z\xc9\xad\x8b\x1a\xa7\xb1Bc\xbe\xe6\xfc\x9c\xa2\xd1\xf8,\xaf\xf1l\xcbz\x0b\xc1N%\xd0\x80\xdf"
    assert isinstance(sig.PHYLUM_MINISIGN_PUBKEY_ED25519, bytes)
    assert sig.PHYLUM_MINISIGN_PUBKEY_ED25519 == expected_key, "The Ed25519 public key should not be changing"



"""Define an implementation for the GitLab CI platform.

GitLab References:
  * https://docs.gitlab.com/ee/ci/
  * https://docs.gitlab.com/ee/ci/docker/using_docker_images.html
  * https://docs.gitlab.com/ee/ci/variables/predefined_variables.html
  * https://docs.gitlab.com/ee/ci/jobs/ci_job_token.html
  * https://docs.gitlab.com/ee/api/notes.html#merge-requests
"""
import os
import subprocess
from argparse import Namespace
from pathlib import Path
from typing import Optional

import requests
from phylum.ci.ci_base import CIBase
from phylum.ci.constants import PHYLUM_HEADER
from phylum.constants import REQ_TIMEOUT


class CIGitLab(CIBase):
    """Provide methods for a GitLab CI environment."""

    def __init__(self, args: Namespace) -> None:
        super().__init__(args)
        self.ci_platform_name = "GitLab CI"

    def _check_prerequisites(self) -> None:
        """Ensure the necessary pre-requisites are met and bail when they aren't.

        These are the current pre-requisites for operating within a GitLab CI Environment:
          * The environment must actually be within GitLab CI
          * A GitLab token providing API access is available
        """
        super()._check_prerequisites()

        if os.getenv("GITLAB_CI") != "true":
            raise SystemExit(" [!] Must be working within the GitLab CI environment")

        # A GitLab token with API access is required to use the API (e.g., to post notes/comments).
        # This can be a personal, project, or group access token...and possibly some other types as well.
        # See the GitLab Token Overview Documentation for info: https://docs.gitlab.com/ee/security/token_overview.html
        gitlab_token = os.getenv("GITLAB_TOKEN")
        if not gitlab_token:
            raise SystemExit(" [!] A GitLab token with API access must be set at `GITLAB_TOKEN`")
        self._gitlab_token = gitlab_token

    @property
    def gitlab_token(self) -> str:
        """Get the GitLab token (e.g., personal, project, group, etc.)."""
        return self._gitlab_token

    @property
    def phylum_label(self) -> str:
        """Get a custom label for use when submitting jobs with `phylum analyze`."""
        mr_iid = os.getenv("CI_MERGE_REQUEST_IID", "unknown-IID")
        mr_title = os.getenv("CI_MERGE_REQUEST_TITLE", "unknown-title")
        label = f"{self.ci_platform_name}_MR#{mr_iid}_{mr_title}"
        label = label.replace(" ", "-")
        return label

    @property
    def common_lockfile_ancestor_commit(self) -> Optional[str]:
        """Find the common lockfile ancestor commit."""
        # Reference: https://docs.gitlab.com/ee/ci/variables/predefined_variables.html
        return os.getenv("CI_MERGE_REQUEST_DIFF_BASE_SHA")

    def _is_lockfile_changed(self, lockfile: Path) -> bool:
        """Predicate for detecting if the given lockfile has changed."""
        mr_src_branch = os.getenv("CI_MERGE_REQUEST_SOURCE_BRANCH_NAME")
        mr_tgt_branch = os.getenv("CI_MERGE_REQUEST_TARGET_BRANCH_NAME")
        mr_diff_base_sha = os.getenv("CI_MERGE_REQUEST_DIFF_BASE_SHA")
        print(f" [+] CI_MERGE_REQUEST_SOURCE_BRANCH_NAME: {mr_src_branch}")
        print(f" [+] CI_MERGE_REQUEST_TARGET_BRANCH_NAME: {mr_tgt_branch}")
        print(f" [+] CI_MERGE_REQUEST_DIFF_BASE_SHA: {mr_diff_base_sha}")

        # Assume no change when there isn't enough information to tell
        if mr_diff_base_sha is None:
            return False

        try:
            # `--exit-code` will make git exit with 1 if there were differences while 0 means no differences.
            # Any other exit code is an error and a reason to re-raise.
            cmd = f"git diff --exit-code --quiet {mr_diff_base_sha} -- {lockfile.resolve()}"
            subprocess.run(cmd.split(), check=True)
            return False
        except subprocess.CalledProcessError as err:
            if err.returncode == 1:
                return True
            print(" [!] Consider changing the `GIT_DEPTH` variable in CI settings to clone/fetch more branch history")
            raise

    def post_output(self) -> None:
        """Post the output of the analysis as a note (comment) on the GitLab CI Merge Request (MR)."""
        # API Reference: https://docs.gitlab.com/ee/api/notes.html#merge-requests
        gitlab_api_v4_root_url = os.getenv("CI_API_V4_URL")
        mr_project_id = os.getenv("CI_MERGE_REQUEST_PROJECT_ID")
        mr_iid = os.getenv("CI_MERGE_REQUEST_IID")
        # This is the same endpoint for listing all MR notes (GET) and creating new ones (POST)
        base_mr_notes_api_endpoint = f"/projects/{mr_project_id}/merge_requests/{mr_iid}/notes"
        url = f"{gitlab_api_v4_root_url}{base_mr_notes_api_endpoint}"
        headers = {"PRIVATE-TOKEN": self.gitlab_token}

        print(f" [*] Getting all current merge request notes with GET URL: {url} ...")
        req = requests.get(url, headers=headers, timeout=REQ_TIMEOUT)
        req.raise_for_status()
        mr_notes = req.json()

        print(" [*] Checking existing merge request notes for existing content to avoid duplication ...")
        if not mr_notes:
            print(" [+] No existing merge request notes found.")
        # NOTE: The API defaults to returning the notes in descending order by the `created_at` field.
        #       Detecting Phylum notes is done simply by looking for those notes that start with a known string value.
        #       We only care about the most recent Phylum note.
        for mr_note in mr_notes:
            if mr_note.get("body", "").lstrip().startswith(PHYLUM_HEADER.strip()):
                print(" [+] The most recently posted Phylum merge request note was found.")
                if mr_note.get("body", "") == self.analysis_output:
                    print(" [+] It contains the same content as the current analysis. Nothing to do.")
                    return
                print(" [+] It does not contain the same content as the current analysis.")
                break

        # If we got here, then the most recent Phylum MR note does not match the current analysis output or
        # there were no Phylum MR notes. Either way, create a new MR note.
        data = {"body": self.analysis_output}
        print(f" [*] Creating new merge request note with POST URL: {url} ...")
        response = requests.post(url, data=data, headers=headers, timeout=REQ_TIMEOUT)
        response.raise_for_status()



"""Define an implementation for when called as a Python pre-commit hook.

This is the case when the `phylum-ci` command is run directly, from the CLI, but not within a CI environment. The
command will be issued by the Python `pre-commit` tool and include extra arguments representing a list of staged files.

References:
  * https://pre-commit.com/index.html#creating-new-hooks
  * https://pre-commit.com/index.html#pre-commit-during-commits
  * https://pre-commit.com/index.html#arguments-pattern-in-hooks
"""
import argparse
import subprocess
from pathlib import Path
from typing import List, Optional

from connect.utils.terminal.markdown import render
from phylum.ci.ci_base import CIBase


class CIPreCommit(CIBase):
    """Provide methods for operating within a pre-commit hook."""

    def __init__(self, args: argparse.Namespace, remainder: List[str]) -> None:
        self.extra_args = remainder
        super().__init__(args)
        self.ci_platform_name = "pre-commit"

    def _check_prerequisites(self) -> None:
        """Ensure the necessary pre-requisites are met and bail when they aren't.

        These are the current pre-requisites for operating within a pre-commit hook:
          * The extra unparsed arguments passed to the CLI represent the staged files, no more and no less
        """
        super()._check_prerequisites()

        cmd = "git diff --cached --name-only".split()
        staged_files = subprocess.run(cmd, check=True, text=True, capture_output=True).stdout.strip().split("\n")
        extra_arg_paths = (Path(extra_arg).resolve() for extra_arg in self.extra_args)

        print(" [*] Checking extra args for valid pre-commit scenarios ...")
        # Allow for a pre-commit config set up to send all staged files to the hook
        if sorted(staged_files) == sorted(self.extra_args):
            print(" [+] The extra args provided exactly match the list of staged files")
        # Allow for a pre-commit config set up to filter the files sent to the hook
        elif all(extra_arg in staged_files for extra_arg in self.extra_args):
            print(" [+] All the extra args are staged files")
        # Allow for cases where the lockfile is included or explicitly specified (e.g., `pre-commit run --all-files`)
        elif self.lockfile in extra_arg_paths:
            print(" [+] The lockfile was included in the extra args")
        # NOTE: There is still the case where the lockfile is "accidentally" included as an extra argument. For example,
        #       `phylum-ci poetry.lock` was used instead of `phylum-ci --lockfile poetry.lock`, which is bad syntax but
        #       nonetheless results in the `CIPreCommit` environment used instead of `CINone`. This is not terrible; it
        #       just might be a slightly confusing corner case. It might be possible to use a library like `psutil` to
        #       acquire the command line from the parent process and inspect it for `pre-commit` usage. That is a
        #       heavyweight solution and one that will not be pursued until the need for it is more clear.
        else:
            print(" [+] No valid pre-commit scenario found. Bailing ...")
            raise SystemExit(f" [!] Unrecognized arguments: {' '.join(self.extra_args)}")

    @property
    def phylum_label(self) -> str:
        """Get a custom label for use when submitting jobs with `phylum analyze`."""
        cmd = "git branch --show-current".split()
        current_branch = subprocess.run(cmd, check=True, text=True, capture_output=True).stdout.strip()

        # This is the unique key that git uses to refer to the blob type data object for the lockfile.
        # Reference: https://git-scm.com/book/en/v2/Git-Internals-Git-Objects
        cmd = f"git hash-object {self.lockfile}".split()
        lockfile_hash_object = subprocess.run(cmd, check=True, text=True, capture_output=True).stdout.strip()
        label = f"{self.ci_platform_name}_{current_branch}_{lockfile_hash_object[:7]}"
        label = label.replace(" ", "-")

        return label

    @property
    def common_lockfile_ancestor_commit(self) -> Optional[str]:
        """Find the common lockfile ancestor commit."""
        cmd = "git rev-parse --verify HEAD".split()
        try:
            common_ancestor_commit = subprocess.run(cmd, check=True, capture_output=True, text=True).stdout.strip()
        except subprocess.CalledProcessError as err:
            print(f" [!] The common lockfile ancestor commit could not be found: {err}")
            common_ancestor_commit = None
        return common_ancestor_commit

    def _is_lockfile_changed(self, lockfile: Path) -> bool:
        """Predicate for detecting if the given lockfile has changed.

        For the case of operating within a pre-commit hook, some assumptions are made:
          * The extra, unparsed, arguments provided to the CLI represent the list of staged files
        """
        staged_files = (Path(staged_file).resolve() for staged_file in self.extra_args)
        return lockfile in staged_files

    def post_output(self) -> None:
        """Post the output of the analysis in the means appropriate for the CI environment."""
        print(f" [+] Analysis output:\n{render(self.analysis_output)}")



"""Define an implementation for the GitHub Actions platform.

GitHub References:
  * https://docs.github.com/en/actions/learn-github-actions/environment-variables#default-environment-variables
  * https://docs.github.com/en/actions/security-guides/automatic-token-authentication
  * https://docs.github.com/en/authentication/keeping-your-account-and-data-secure/creating-a-personal-access-token
  * https://docs.github.com/en/developers/apps/building-oauth-apps/scopes-for-oauth-apps#available-scopes
  * https://docs.github.com/en/developers/webhooks-and-events/webhooks/webhook-events-and-payloads#pull_request
  * https://docs.github.com/en/rest/overview/resources-in-the-rest-api
  * https://docs.github.com/en/rest/pulls/comments
  * https://docs.github.com/en/rest/guides/working-with-comments#pull-request-comments
"""
import json
import os
import subprocess
from argparse import Namespace
from pathlib import Path
from typing import Optional

import requests
from phylum.ci.ci_base import CIBase
from phylum.ci.constants import PHYLUM_HEADER
from phylum.constants import REQ_TIMEOUT


class CIGitHub(CIBase):
    """Provide methods for a GitHub Actions environment."""

    def __init__(self, args: Namespace) -> None:
        # This is the recommended workaround for container actions, to avoid the `unsafe repository` error.
        # It is added before super().__init__(args) so that lockfile change detection will be set properly.
        # See https://github.com/actions/checkout/issues/766 (git CVE-2022-24765) for more detail.
        github_workspace = os.getenv("GITHUB_WORKSPACE", "/github/workspace")
        cmd = f"git config --global --add safe.directory {github_workspace}"
        subprocess.run(cmd.split(), check=True)

        super().__init__(args)
        self.ci_platform_name = "GitHub Actions"

    def _check_prerequisites(self) -> None:
        """Ensure the necessary pre-requisites are met and bail when they aren't.

        These are the current pre-requisites for operating within a GitHub Actions Environment:
          * The environment must actually be within GitHub Actions
          * A GitHub token providing `issues` API access is available
          * `pull_request` webhook event payload is available
        """
        super()._check_prerequisites()

        if os.getenv("GITHUB_ACTIONS") != "true":
            raise SystemExit(" [!] Must be working within the GitHub Actions environment")

        # A GitHub token with API access is required to use the API (e.g., to post comments). This can be the default
        # `GITHUB_TOKEN` provided automatically at the start of each workflow run or a personal access token (PAT).
        # A `GITHUB_TOKEN` needs at least write access for `pull-requests` scope (even though the `issues` API is used).
        # A PAT needs the `repo` scope or minimally the `public_repo` scope if private repositories are not used.
        # See the GitHub Token Documentation for more info:
        # https://docs.github.com/en/actions/security-guides/automatic-token-authentication
        # https://docs.github.com/en/authentication/keeping-your-account-and-data-secure/creating-a-personal-access-token
        # https://docs.github.com/en/developers/apps/building-oauth-apps/scopes-for-oauth-apps#available-scopes
        github_token = os.getenv("GITHUB_TOKEN")
        if not github_token:
            raise SystemExit(" [!] A GitHub token with API access must be set at `GITHUB_TOKEN` environment variable")
        self._github_token = github_token

        # Unfortunately, there's not always a simple default environment variable that contains the desired information.
        # Instead, the full event webhook payload can be used to obtain the information. Reference:
        # https://docs.github.com/en/developers/webhooks-and-events/webhooks/webhook-events-and-payloads#pull_request
        if os.getenv("GITHUB_EVENT_NAME") != "pull_request":
            raise SystemExit(" [!] The workflow event must be `pull_request`")
        github_event_path = os.getenv("GITHUB_EVENT_PATH")
        if github_event_path is None:
            raise SystemExit(" [!] Could not read the `GITHUB_EVENT_PATH` environment variable")
        with open(github_event_path, encoding="utf-8") as f:
            pr_event = json.load(f)
        self._pr_event = pr_event

    @property
    def github_token(self) -> str:
        """Get the default `GITHUB_TOKEN` or custom Personal Access Token (PAT) in use."""
        return self._github_token

    @property
    def pr_event(self) -> dict:
        """Get the `pull_request` webhook event payload."""
        return self._pr_event

    @property
    def phylum_label(self) -> str:
        """Get a custom label for use when submitting jobs with `phylum analyze`."""
        pr_number = self.pr_event.get("pull_request", {}).get("number", "unknown-number")
        pr_src_branch = os.getenv("GITHUB_HEAD_REF", "unknown-ref")
        label = f"{self.ci_platform_name}_PR#{pr_number}_{pr_src_branch}"
        label = label.replace(" ", "-")
        return label

    @property
    def common_lockfile_ancestor_commit(self) -> Optional[str]:
        """Find the common lockfile ancestor commit."""
        return self.pr_event.get("pull_request", {}).get("base", {}).get("sha")

    def _is_lockfile_changed(self, lockfile: Path) -> bool:
        """Predicate for detecting if the given lockfile has changed."""
        pr_src_branch = os.getenv("GITHUB_HEAD_REF")
        pr_tgt_branch = os.getenv("GITHUB_BASE_REF")
        pr_base_sha = self.common_lockfile_ancestor_commit
        print(f" [+] GITHUB_HEAD_REF: {pr_src_branch}")
        print(f" [+] GITHUB_BASE_REF: {pr_tgt_branch}")
        print(f" [+] PR base SHA: {pr_base_sha}")

        # Assume no change when there isn't enough information to tell
        if pr_base_sha is None:
            return False

        try:
            # `--exit-code` will make git exit with 1 if there were differences while 0 means no differences.
            # Any other exit code is an error and a reason to re-raise.
            cmd = f"git diff --exit-code --quiet {pr_base_sha} -- {lockfile.resolve()}"
            subprocess.run(cmd.split(), check=True)
            return False
        except subprocess.CalledProcessError as err:
            if err.returncode == 1:
                return True
            print(" [!] Consider changing the `fetch_depth` input during checkout to fetch more branch history")
            raise

    def post_output(self) -> None:
        """Post the output of the analysis as a comment on the GitHub Pull Request (PR)."""
        # API Reference: https://docs.github.com/en/rest/issues/comments
        # This is the same endpoint for listing all PR comments (GET) and creating new ones (POST)
        comments_url = self.pr_event.get("pull_request", {}).get("comments_url")

        headers = {
            "Accept": "application/vnd.github.v3+json",
            "Authorization": f"token {self.github_token}",
        }

        query_params = {"per_page": 100}
        print(f" [*] Getting all current pull request comments with GET URL: {comments_url} ...")
        req = requests.get(comments_url, headers=headers, params=query_params, timeout=REQ_TIMEOUT)
        req.raise_for_status()
        pr_comments = req.json()

        print(" [*] Checking pull request comments for existing content to avoid duplication ...")
        if not pr_comments:
            print(" [+] No existing pull request comments found.")
        # NOTE: The API call returns the comments in ascending order by ID...thus the need to reverse the list.
        #       Detecting Phylum comments is done simply by looking for those that start with a known string value.
        #       We only care about the most recent Phylum comment.
        for pr_comment in reversed(pr_comments):
            if pr_comment.get("body", "").lstrip().startswith(PHYLUM_HEADER.strip()):
                print(" [+] The most recently posted Phylum pull request comment was found.")
                if pr_comment.get("body", "") == self.analysis_output:
                    print(" [+] It contains the same content as the current analysis. Nothing to do.")
                    return
                print(" [+] It does not contain the same content as the current analysis.")
                break

        # If we got here, then the most recent Phylum PR comment does not match the current analysis output or
        # there were no Phylum PR comments. Either way, create a new PR comment.
        body_params = {"body": self.analysis_output}
        print(f" [*] Creating new pull request comment with POST URL: {comments_url} ...")
        response = requests.post(comments_url, headers=headers, json=body_params, timeout=REQ_TIMEOUT)
        response.raise_for_status()



"""Console script for phylum-init."""
import argparse
import os
import pathlib
import platform
import shutil
import subprocess
import sys
import tempfile
import zipfile
from functools import lru_cache
from pathlib import Path
from typing import List, Optional, Tuple

import requests
from packaging.utils import canonicalize_version
from packaging.version import InvalidVersion, Version
from phylum import __version__
from phylum.constants import (
    MIN_SUPPORTED_CLI_VERSION,
    REQ_TIMEOUT,
    SUPPORTED_ARCHES,
    SUPPORTED_PLATFORMS,
    TOKEN_ENVVAR_NAME,
)
from phylum.init import SCRIPT_NAME
from phylum.init.sig import verify_minisig
from ruamel.yaml import YAML


def use_legacy_paths(version):
    """Predicate to specify whether legacy paths should be used for a given version.

    The Phylum config and binary paths changed following the v2.2.0 release, to adhere to the XDG Base Directory Spec.
    Reference: https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html
    """
    return Version(canonicalize_version(version)) <= Version("v2.2.0")


def get_phylum_settings_path(version):
    """Get the Phylum settings path based on a provided version."""
    home_dir = pathlib.Path.home()
    version = version_check(version)

    config_home_path = os.getenv("XDG_CONFIG_HOME")
    if not config_home_path:
        config_home_path = home_dir / ".config"

    phylum_config_path = pathlib.Path(config_home_path) / "phylum" / "settings.yaml"
    if use_legacy_paths(version):
        phylum_config_path = home_dir / ".phylum" / "settings.yaml"

    return phylum_config_path


def get_expected_phylum_bin_path(version):
    """Get the expected path to the Phylum CLI binary based on a provided version."""
    home_dir = pathlib.Path.home()
    version = version_check(version)

    phylum_bin_path = home_dir / ".local" / "bin" / "phylum"
    if use_legacy_paths(version):
        phylum_bin_path = home_dir / ".phylum" / "phylum"

    return phylum_bin_path


def get_phylum_cli_version(cli_path: Path) -> str:
    """Get the version of the installed and active Phylum CLI and return it."""
    cmd = f"{cli_path} --version"
    version = subprocess.run(cmd.split(), check=True, capture_output=True, text=True).stdout.strip().lower()

    # Starting with Python 3.9, the str.removeprefix() method was introduced to do this same thing
    prefix = "phylum "
    prefix_len = len(prefix)
    if version.startswith(prefix):
        version = version[prefix_len:]

    return version


def get_phylum_bin_path(version: str = None) -> Tuple[Optional[Path], Optional[str]]:
    """Get the current path and corresponding version to the Phylum CLI binary and return them.

    Provide a CLI version as a fallback method for looking on an explicit path,
    based on the expected path for that version.
    """
    # Look for `phylum` on the PATH first
    which_cli_path = shutil.which("phylum")

    if which_cli_path is None and version is not None:
        # Maybe `phylum` is installed already but not on the PATH or maybe the PATH has not been updated in this
        # context. Look in the specific location expected by the provided version.
        expected_cli_path = get_expected_phylum_bin_path(version)
        which_cli_path = shutil.which("phylum", path=expected_cli_path)

    if which_cli_path is None:
        return (None, None)

    cli_path = Path(which_cli_path)
    cli_version = get_phylum_cli_version(cli_path)
    return cli_path, cli_version


def get_latest_version():
    """Get the "latest" version programmatically and return it."""
    # API Reference: https://docs.github.com/en/rest/releases/releases#get-the-latest-release
    github_api_url = "https://api.github.com/repos/phylum-dev/cli/releases/latest"

    headers = {"Accept": "application/vnd.github+json"}
    req = requests.get(github_api_url, headers=headers, timeout=REQ_TIMEOUT)
    req.raise_for_status()
    req_json = req.json()

    # The "name" entry stores the GitHub Release name, which could be set to something other than the version.
    # Using the "tag_name" entry is better since the tags are much more tightly coupled with the release version.
    latest_version = req_json.get("tag_name")

    return latest_version


@lru_cache(maxsize=1)
def supported_releases() -> List[str]:
    """Get the most recent supported releases programmatically and return them."""
    # API Reference: https://docs.github.com/en/rest/releases/releases#list-releases
    github_api_url = "https://api.github.com/repos/phylum-dev/cli/releases"

    headers = {"Accept": "application/vnd.github+json"}
    query_params = {"per_page": 100}
    req = requests.get(github_api_url, headers=headers, params=query_params, timeout=REQ_TIMEOUT)
    req.raise_for_status()
    req_json = req.json()

    # The "name" entry stores the GitHub Release name, which could be set to something other than the version.
    # Using the "tag_name" entry is better since the tags are much more tightly coupled with the release version.
    releases = [rel.get("tag_name") for rel in req_json if is_supported_version(rel.get("tag_name", "0.0.0"))]

    return releases


def is_supported_version(version: str) -> bool:
    """Predicate for determining if a given version is supported."""
    try:
        provided_version = Version(canonicalize_version(version))
        min_supported_version = Version(MIN_SUPPORTED_CLI_VERSION)
    except InvalidVersion as err:
        raise ValueError("An invalid version was provided") from err

    return provided_version >= min_supported_version


def supported_targets(release_tag: str) -> List[str]:
    """Get the supported Rust target triples programmatically for a given release tag and return them.

    Targets are identified by their "target triple" which is the string to inform the compiler what kind of output
    should be produced. A target triple consists of three strings separated by a hyphen, with a possible fourth string
    at the end preceded by a hyphen. The first is the architecture, the second is the "vendor", the third is the OS
    type, and the optional fourth is environment type.

    References:
      * https://doc.rust-lang.org/nightly/rustc/platform-support.html
      * https://rust-lang.github.io/rfcs/0131-target-specification.html
    """
    if release_tag not in supported_releases():
        raise SystemExit(f" [!] Unsupported version: {release_tag}")

    # API Reference: https://docs.github.com/en/rest/releases/releases#get-a-release-by-tag-name
    github_api_url = f"https://api.github.com/repos/phylum-dev/cli/releases/tags/{release_tag}"

    headers = {"Accept": "application/vnd.github+json"}
    req = requests.get(github_api_url, headers=headers, timeout=REQ_TIMEOUT)
    req.raise_for_status()
    req_json = req.json()

    assets = req_json.get("assets", [])
    targets: List[str] = []
    prefix, suffix = "phylum-", ".zip"
    for asset in assets:
        name = asset.get("name", "")
        if name.startswith(prefix) and name.endswith(suffix):
            target = name.replace(prefix, "").replace(suffix, "")
            targets.append(target)

    return list(set(targets))


def version_check(version):
    """Check a given version for validity and return a normalized form of it."""
    if version == "latest":
        version = get_latest_version()

    version = version.lower()
    if not version.startswith("v"):
        version = f"v{version}"

    supported_versions = supported_releases()
    if version not in supported_versions:
        releases = ", ".join(supported_versions)
        raise argparse.ArgumentTypeError(f"version must be from a supported release: {releases}")

    return version


def get_target_triple():
    """Get the "target triple" from the current system and return it."""
    arch = SUPPORTED_ARCHES.get(platform.uname().machine.lower(), "unknown")
    plat = SUPPORTED_PLATFORMS.get(platform.uname().system.lower(), "unknown")
    return f"{arch}-{plat}"


def save_file_from_url(url, path):
    """Save a file from a given URL to a local file path, in binary mode."""
    print(f" [*] Getting {url} file ...", end="")
    req = requests.get(url, timeout=REQ_TIMEOUT)
    req.raise_for_status()
    print("Done")

    print(f" [*] Saving {url} file to {path} ...", end="")
    with open(path, "wb") as f:
        f.write(req.content)
    print("Done")


def get_archive_url(tag_name, archive_name):
    """Craft an archive download URL from a given tag name and archive name."""
    # Reference: https://docs.github.com/en/rest/releases/releases#get-a-release-by-tag-name
    github_base_uri = "https://github.com/phylum-dev/cli/releases"
    archive_url = f"{github_base_uri}/download/{tag_name}/{archive_name}"
    return archive_url


def is_token_set(phylum_settings_path, token=None):
    """Check if any token is already set in the given CLI configuration file.

    Optionally, check if a specific given `token` is set.
    """
    try:
        settings_data = phylum_settings_path.read_text(encoding="utf-8")
    except FileNotFoundError:
        return False

    yaml = YAML()
    settings_dict = yaml.load(settings_data)
    configured_token = settings_dict.get("auth_info", {}).get("offline_access")

    if configured_token is None:
        return False
    if token is not None:
        if token != configured_token:
            return False

    return True


def process_token_option(args):
    """Process the token option as parsed from the arguments."""
    phylum_settings_path = get_phylum_settings_path(args.version)

    # The token option takes precedence over the Phylum API key environment variable.
    token = os.getenv(TOKEN_ENVVAR_NAME)
    if args.token is not None:
        token = args.token

    if token:
        print(f" [+] Phylum token supplied as an option or `{TOKEN_ENVVAR_NAME}` environment variable")
        if is_token_set(phylum_settings_path):
            print(" [+] An existing token is already set")
            if is_token_set(phylum_settings_path, token=token):
                print(" [+] Supplied token matches existing token")
            else:
                print(" [!] Supplied token will be used to overwrite the existing token")
        else:
            print(" [+] No existing token exists. Supplied token will be used.")
    else:
        print(f" [+] Phylum token NOT supplied as option or `{TOKEN_ENVVAR_NAME}` environment variable")
        if is_token_set(phylum_settings_path):
            print(" [+] Existing token found. It will be used without modification.")
        else:
            print(" [!] Existing token not found. Use `phylum auth login` or `phylum auth register` command to set it.")

    if token and not is_token_set(phylum_settings_path, token=token):
        setup_token(token, args)


def setup_token(token, args):
    """Setup the CLI credentials with a provided token and path to phylum binary."""
    phylum_bin_path = get_expected_phylum_bin_path(args.version)
    phylum_settings_path = get_phylum_settings_path(args.version)

    # The phylum CLI settings.yaml file won't exist upon initial install
    # but running a command will trigger the CLI to generate it
    if not phylum_settings_path.exists():
        cmd = f"{phylum_bin_path} version".split()
        subprocess.run(cmd, check=True)

    yaml = YAML()
    settings_dict = yaml.load(phylum_settings_path.read_text(encoding="utf-8"))
    settings_dict.setdefault("auth_info", {})
    settings_dict["auth_info"]["offline_access"] = token
    with open(phylum_settings_path, "w", encoding="utf-8") as f:
        yaml.dump(settings_dict, f)

    # Check that the token was setup correctly by using it to display the current auth status
    cmd = f"{phylum_bin_path} auth status".split()
    subprocess.run(cmd, check=True)


def get_args(args=None):
    """Get the arguments from the command line and return them."""
    parser = argparse.ArgumentParser(
        prog=SCRIPT_NAME,
        description="Fetch and install the Phylum CLI",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"{SCRIPT_NAME} {__version__}",
    )
    parser.add_argument(
        "-r",
        "--phylum-release",
        dest="version",
        default="latest",
        type=version_check,
        help="""The version of the Phylum CLI to install. Can be specified as `latest` or a specific tagged release,
            with or without the leading `v`.""",
    )
    parser.add_argument(
        "-t",
        "--target",
        default=get_target_triple(),
        help="The target platform type where the CLI will be installed.",
    )
    parser.add_argument(
        "-k",
        "--phylum-token",
        dest="token",
        help=f"""Phylum user token. Can also specify this option's value by setting the `{TOKEN_ENVVAR_NAME}`
            environment variable. The value specified with this option takes precedence when both are provided.
            Leave this option and it's related environment variable unspecified to either (1) use an existing token
            already set in the Phylum config file or (2) to manually populate the token with a `phylum auth login` or
            `phylum auth register` command after install.""",
    )

    list_group = parser.add_mutually_exclusive_group()
    list_group.add_argument(
        "--list-releases",
        action="store_true",
        help="List the Phylum CLI releases available to install.",
    )
    list_group.add_argument(
        "--list-targets",
        action="store_true",
        help="List the target platform types available for installing a given Phylum CLI release.",
    )

    return parser.parse_args(args=args)


def main(args=None):
    """Main entrypoint."""
    args = get_args(args=args)

    if args.list_releases:
        print("Looking up supported releases ...")
        releases = ", ".join(supported_releases())
        print(f"Supported releases: {releases}")
        return 0

    tag_name = args.version
    supported_target_triples = supported_targets(tag_name)
    if args.list_targets:
        print(f"Looking up supported targets for release {tag_name} ...")
        targets = ", ".join(supported_target_triples)
        print(f"Supported targets for release {tag_name}: {targets}")
        return 0

    target_triple = args.target
    if target_triple not in supported_target_triples:
        raise SystemExit(f" [!] The identified target triple `{target_triple}` is not supported for release {tag_name}")

    archive_name = f"phylum-{target_triple}.zip"
    minisig_name = f"{archive_name}.minisig"
    archive_url = get_archive_url(tag_name, archive_name)
    minisig_url = f"{archive_url}.minisig"
    phylum_bin_path = get_expected_phylum_bin_path(tag_name)

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = pathlib.Path(temp_dir)
        archive_path = temp_dir_path / archive_name
        minisig_path = temp_dir_path / minisig_name

        save_file_from_url(archive_url, archive_path)
        save_file_from_url(minisig_url, minisig_path)

        verify_minisig(archive_path, minisig_path)

        with zipfile.ZipFile(archive_path, mode="r") as zip_file:
            if zip_file.testzip() is not None:
                raise zipfile.BadZipFile(f"There was a bad file in the zip archive {archive_name}")
            extracted_dir = temp_dir_path / f"phylum-{target_triple}"
            zip_file.extractall(path=temp_dir)

        cmd = "sh install.sh".split()
        subprocess.run(cmd, check=True, cwd=extracted_dir)

    process_token_option(args)

    # Check to ensure everything is working
    cmd = f"{phylum_bin_path} --help".split()
    subprocess.run(cmd, check=True)

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""Place test package constants here."""
import pathlib

import tomli

HERE = pathlib.Path(__file__).resolve().parent
PROJECT_ROOT = HERE.parent
PYPROJECT_TOML_PATH = PROJECT_ROOT / "pyproject.toml"

with open(PYPROJECT_TOML_PATH, "rb") as f:
    PYPROJECT = tomli.load(f)



"""Define an implementation for when there is no active CI platform.

This is the case when the `phylum-ci` command is run directly, from the CLI, but not within a CI environment.
This might be useful for running locally.
This is also the fallback implementation to use when no known CI platform is detected.
"""
import argparse
import subprocess
from pathlib import Path
from typing import Optional

from connect.utils.terminal.markdown import render
from phylum.ci.ci_base import CIBase


def git_remote() -> str:
    """Get the git remote and return it.

    This function is limited in that it will only work when there is a single remote defined.
    A RuntimeError exception will be raised when there is not exactly one remote.
    """
    cmd = "git remote"
    remotes = subprocess.run(cmd.split(), check=True, text=True, capture_output=True).stdout.splitlines()
    if not remotes:
        raise RuntimeError("No git remotes configured")
    if len(remotes) > 1:
        raise RuntimeError("Only one git remote is supported at this time")
    remote = remotes[0]
    return remote


class CINone(CIBase):
    """Provide methods for operating outside of a known CI environment."""

    def __init__(self, args: argparse.Namespace) -> None:
        super().__init__(args)
        self.ci_platform_name = "No CI"

    def _check_prerequisites(self) -> None:
        """Ensure the necessary pre-requisites are met and bail when they aren't.

        These are the current pre-requisites for when no CI environments/platforms is detected:
          * Run the script from the root of a git repository
        """
        super()._check_prerequisites()

        git_dir = Path.cwd() / ".git"
        if git_dir.is_dir():
            print(" [+] Existing `.git` directory was found at the current working directory")
        else:
            raise SystemExit(" [!] This script expects to be run from the top level of a `git` repository")

    @property
    def phylum_label(self) -> str:
        """Get a custom label for use when submitting jobs with `phylum analyze`."""
        cmd = "git branch --show-current".split()
        current_branch = subprocess.run(cmd, check=True, text=True, capture_output=True).stdout.strip()

        # This is the unique key that git uses to refer to the blob type data object for the lockfile.
        # Reference: https://git-scm.com/book/en/v2/Git-Internals-Git-Objects
        cmd = f"git hash-object {self.lockfile}".split()
        lockfile_hash_object = subprocess.run(cmd, check=True, text=True, capture_output=True).stdout.strip()
        label = f"{self.ci_platform_name}_{current_branch}_{lockfile_hash_object[:7]}"
        label = label.replace(" ", "-")

        return label

    @property
    def common_lockfile_ancestor_commit(self) -> Optional[str]:
        """Find the common lockfile ancestor commit."""
        remote = git_remote()
        cmd = f"git merge-base HEAD refs/remotes/{remote}/HEAD".split()
        try:
            common_ancestor_commit = subprocess.run(cmd, check=True, capture_output=True, text=True).stdout.strip()
        except subprocess.CalledProcessError as err:
            print(f" [!] The common lockfile ancestor commit could not be found: {err}")
            common_ancestor_commit = None
        return common_ancestor_commit

    def _is_lockfile_changed(self, lockfile: Path) -> bool:
        """Predicate for detecting if the given lockfile has changed.

        For the case of operating outside of a CI platform, some assumptions are made:
          * There is only one remote configured for the repository
          * The diff is comparing against the remote and not another ref
          * The diff is comparing by using the files at the current HEAD

        The usefulness of this approach is limited in that lockfile changes must already be committed to be detected.

        References:
        https://git-scm.com/docs/git-diff#Documentation/git-diff.txt-emgitdiffemltoptionsgtltcommitgtltcommitgt--ltpathgt82308203
        """
        remote = git_remote()
        cmd = f"git diff --exit-code --quiet refs/remotes/{remote}/HEAD... -- {lockfile.resolve()}".split()
        try:
            # `--exit-code` will make git exit with with 1 if there were differences while 0 means no differences.
            # Any other exit code is an error and a reason to re-raise.
            subprocess.run(cmd, check=True)
            return False
        except subprocess.CalledProcessError as err:
            if err.returncode == 1:
                return True
            raise

    def post_output(self) -> None:
        """Post the output of the analysis in the means appropriate for the CI environment."""
        print(f" [+] Analysis output:\n{render(self.analysis_output)}")



"""Define a base environment for CI platforms.

The "base" environment is one that makes use of the CLI directly and is not necessarily part of a continuous
integration (CI) environment. Common functionality is provided where possible and CI specific features are
designated as abstract methods to be defined in specific CI environments.
"""
import json
import os
import shutil
import subprocess
import tempfile
import urllib.parse
from abc import ABC, abstractmethod
from argparse import Namespace
from pathlib import Path
from typing import List, Optional, Tuple

from packaging.version import Version
from phylum.ci.common import PackageDescriptor, Packages, ProjectThresholdInfo, ReturnCode, RiskDomain
from phylum.ci.constants import (
    FAILED_COMMENT,
    INCOMPLETE_COMMENT_TEMPLATE,
    INCOMPLETE_WITH_FAILURE_COMMENT_TEMPLATE,
    PROJECT_THRESHOLD_OPTIONS,
    SUCCESS_COMMENT,
)
from phylum.constants import SUPPORTED_LOCKFILES, TOKEN_ENVVAR_NAME
from phylum.init.cli import get_phylum_bin_path
from phylum.init.cli import main as phylum_init
from ruamel.yaml import YAML


def detect_lockfile() -> Optional[Path]:
    """Detect the lockfile in use and return it.

    This method has some assumptions:
      * It is called from the root of a repository
      * There is only one lockfile allowed
    """
    cwd = Path.cwd()
    lockfiles_in_cwd = [path.resolve() for lockfile_glob in SUPPORTED_LOCKFILES for path in cwd.glob(lockfile_glob)]
    lockfiles_in_cwd = [path for path in lockfiles_in_cwd if path.stat().st_size]
    if not lockfiles_in_cwd:
        return None
    if len(lockfiles_in_cwd) > 1:
        lockfiles = ", ".join(str(lockfile) for lockfile in lockfiles_in_cwd)
        print(f" [!] Multiple lockfiles detected: {lockfiles}")
        raise SystemExit(" [!] Only one lockfile is supported at this time. Consider specifying it with `--lockfile`.")
    lockfile = lockfiles_in_cwd[0]
    return lockfile


class CIBase(ABC):
    """Provide methods for a basic CI environment."""

    @abstractmethod
    def __init__(self, args: Namespace) -> None:
        """Initialize the base class object.

        Each child class is expected to at least:
          * define `self.ci_platform_name`
          * call `super().__init__(args)`
        """
        self.args = args
        self._phylum_project_file = Path.cwd().joinpath(".phylum_project").resolve()

        # The lockfile specified as a script argument will be used, if provided.
        # Otherwise, an attempt will be made to automatically detect the lockfile.
        provided_lockfile: Path = args.lockfile
        if provided_lockfile and provided_lockfile.exists() and provided_lockfile.stat().st_size:
            self._lockfile = provided_lockfile.resolve()
        else:
            detected_lockfile = detect_lockfile()
            if detected_lockfile:
                self._lockfile = detected_lockfile
            else:
                raise SystemExit(
                    " [!] A lockfile is required and was not detected. Consider specifying one with `--lockfile`."
                )

        # Ensure all pre-requisites are met and bail at the earliest opportunity when they aren't
        self._check_prerequisites()
        print(" [+] All pre-requisites met")

        self._cli_path: Optional[Path] = None
        self.gbl_failed = False
        self.gbl_incomplete = False
        self.incomplete_pkgs: Packages = []
        self._analysis_output = "No analysis output yet"
        self.ci_platform_name = "Unknown"

        # The token option takes precedence over the Phylum API key environment variable.
        token = os.getenv(TOKEN_ENVVAR_NAME)
        if args.token is not None:
            token = args.token
            os.environ[TOKEN_ENVVAR_NAME] = args.token
        self.args.token = token

        self._lockfile_changed = self._is_lockfile_changed(self.lockfile)

    @property
    def lockfile(self) -> Path:
        """Get the package lockfile."""
        return self._lockfile

    @property
    def is_lockfile_changed(self) -> bool:
        """Get the lockfile's modification status."""
        return self._lockfile_changed

    @property
    def phylum_project_file(self) -> Path:
        """Get the path to the Phylum project file `.phylum_project`."""
        return self._phylum_project_file

    @property
    def cli_path(self) -> Optional[Path]:
        """Get the path to the Phylum CLI binary."""
        return self._cli_path

    @property
    def analysis_output(self) -> str:
        """Get the output from the overall analysis, in markdown format."""
        return self._analysis_output

    @property
    @abstractmethod
    def phylum_label(self) -> str:
        """Get a custom label for use when submitting jobs with `phylum analyze`.

        Each CI platform/environment has unique ways of referencing events, PRs, branches, etc.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def common_lockfile_ancestor_commit(self) -> Optional[str]:
        """Find the common lockfile ancestor commit.

        When found, it should be returned as a string of the SHA1 sum representing the commit.
        When it can't be found (or there is an error), `None` should be returned.
        """
        raise NotImplementedError()

    @abstractmethod
    def _is_lockfile_changed(self, lockfile: Path) -> bool:
        """Predicate for detecting if the given lockfile has changed."""
        raise NotImplementedError()

    @abstractmethod
    def _check_prerequisites(self) -> None:
        """Ensure the necessary pre-requisites are met and bail when they aren't.

        The current pre-requisites for *all* CI environments/platforms are:
          * A `.phylum_project` file exists at the working directory
          * Phylum CLI v3.3.0+, to make use of the `parse` command
          * Have `git` installed and available for use on the PATH
        """
        print(" [+] Confirming pre-requisites ...")

        if self.phylum_project_file.exists():
            print(f" [+] Existing `.phylum_project` file found at: {self.phylum_project_file}")
        else:
            raise SystemExit(" [!] The `.phylum_project` file was not found at the current working directory")

        # The `parse` command was available in the pre-releases, but it makes the
        # error message cleaner to only mention the release version.
        if Version(self.args.version) < Version("v3.3.0-rc1"):
            raise SystemExit(" [!] The CLI version must be at least v3.3.0")

        if shutil.which("git"):
            print(" [+] `git` binary found on the PATH")
        else:
            raise SystemExit(" [!] `git` is required to be installed and available on the PATH")

    @abstractmethod
    def post_output(self) -> None:
        """Post the output of the analysis in the means appropriate for the CI environment.

        Output in the form of comments on a pull/merge request should be unique and not added multiple times as
        the review changes but the lock file doesn't.
        """
        raise NotImplementedError()

    # TODO: Use the `@functools.cached_property` decorator, introduced in Python 3.8, to avoid computing more than once.
    #       https://github.com/phylum-dev/phylum-ci/issues/18
    @property
    def current_lockfile_packages(self) -> Packages:
        """Get the current lockfile packages.

        This property expects the Phylum CLI path to be known and will raise `SystemExit` when that path is unknown.
        """
        if not self.cli_path:
            raise SystemExit(" [!] Phylum CLI path is unknown. Try using the `init_cli` method first.")
        try:
            cmd = f"{self.cli_path} parse {self.lockfile}".split()
            parse_result = subprocess.run(cmd, check=True, capture_output=True, text=True).stdout.strip()
        except subprocess.CalledProcessError as err:
            print(f" [!] There was an error running the command: {' '.join(err.cmd)}")
            print(f" [!] stderr:\n{err.stderr}")
            raise SystemExit(f" [!] Is {self.lockfile} valid? If so, please report this as a bug.") from err
        parsed_pkgs = json.loads(parse_result)
        curr_lockfile_packages = [PackageDescriptor(**pkg) for pkg in parsed_pkgs]
        return curr_lockfile_packages

    @property
    def previous_lockfile_object(self) -> Optional[str]:
        """Get the previous git object for the lockfile."""
        if not self.common_lockfile_ancestor_commit:
            return None
        try:
            cmd_line = f"git rev-parse --verify {self.common_lockfile_ancestor_commit}:{self.lockfile.name}".split()
            prev_lockfile_object = subprocess.run(cmd_line, check=True, capture_output=True, text=True).stdout.strip()
        except subprocess.CalledProcessError as err:
            # There could be a true error, but the working assumption when here is a previous version does not exist
            print(f" [?] There *may* be an issue with the attempt to get the previous lockfile object: {err}")
            print(f" [?] stderr: {err.stderr}")
            print(" [+] Assuming a previous lockfile version does not exist ...")
            prev_lockfile_object = None
        return prev_lockfile_object

    def get_previous_lockfile_packages(self, prev_lockfile_object: str) -> Packages:
        """Get the previous lockfile packages from the corresponding git object and return them.

        Expects the Phylum CLI path to be known and will raise `SystemExit` when that path is unknown.
        """
        if not self.cli_path:
            raise SystemExit(" [!] Phylum CLI path is unknown. Try using the `init_cli` method first.")

        with tempfile.NamedTemporaryFile(mode="w+") as prev_lockfile_fd:
            try:
                cmd = f"git cat-file blob {prev_lockfile_object}"
                prev_lockfile_contents = subprocess.run(cmd.split(), check=True, capture_output=True, text=True).stdout
                prev_lockfile_fd.write(prev_lockfile_contents)
                prev_lockfile_fd.flush()
            except subprocess.CalledProcessError as err:
                print(f" [!] There was an error running the command: {' '.join(err.cmd)}")
                print(f" [!] stdout:\n{err.stdout}")
                print(f" [!] stderr:\n{err.stderr}")
                print(" [!] Due to error, assuming no previous lockfile packages. Please report this as a bug.")
                return []
            try:
                cmd = f"{self.cli_path} parse {prev_lockfile_fd.name}"
                parse_result = subprocess.run(cmd.split(), check=True, capture_output=True, text=True).stdout.strip()
            except subprocess.CalledProcessError as err:
                print(f" [!] There was an error running the command: {' '.join(err.cmd)}")
                print(f" [!] stdout:\n{err.stdout}")
                print(f" [!] stderr:\n{err.stderr}")
                print(" [!] Due to error, assuming no previous lockfile packages. Please report this as a bug.")
                return []

        parsed_pkgs = json.loads(parse_result)
        prev_lockfile_packages = [PackageDescriptor(**pkg) for pkg in parsed_pkgs]
        return prev_lockfile_packages

    def get_new_deps(self) -> Packages:
        """Get the new dependencies added to the lockfile and return them."""
        curr_lockfile_packages = self.current_lockfile_packages

        prev_lockfile_object = self.previous_lockfile_object
        if not prev_lockfile_object:
            print(" [+] No previous lockfile object found. Assuming all packages in the current lockfile are new.")
            return curr_lockfile_packages

        prev_lockfile_packages = self.get_previous_lockfile_packages(prev_lockfile_object)

        prev_pkg_set = set(prev_lockfile_packages)
        curr_pkg_set = set(curr_lockfile_packages)
        new_deps = curr_pkg_set.difference(prev_pkg_set)
        print(f" [+] New dependencies: {new_deps}")

        return list(new_deps)

    def init_cli(self) -> None:
        """Check for an existing Phylum CLI install, install it if needed, and set the path class instance variable."""
        specified_version = self.args.version
        # fmt: off
        install_args = [
            "--phylum-release", specified_version,
            "--target", self.args.target,
            "--phylum-token", self.args.token,
        ]
        # fmt: on
        cli_path, cli_version = get_phylum_bin_path(version=specified_version)
        if cli_path is None:
            print(f" [+] Existing Phylum CLI instance not found. Installing version `{specified_version}` ...")
            phylum_init(install_args)
        else:
            print(f" [+] Existing Phylum CLI instance found: {cli_version} at {cli_path}")
            if cli_version != specified_version:
                print(f" [+] Existing version {cli_version} does not match the specified version {specified_version}")
                if self.args.force_install:
                    print(f" [*] Installing Phylum CLI version {specified_version} ...")
                    phylum_init(install_args)
                else:
                    print(" [+] Attempting to use existing version ...")
                    if Version(str(cli_version)) < Version("v3.2.0"):
                        raise SystemExit(" [!] The existing CLI version must be greater than v3.2.0")
                    print(" [+] Version checks succeeded. Using existing version.")

        cli_path, cli_version = get_phylum_bin_path(version=specified_version)
        print(f" [+] Using Phylum CLI instance: {cli_version} at {str(cli_path)}")

        self._cli_path = cli_path

        # Exit condition: a Phylum API key should be in place or available at this point.
        # Ensure stdout is piped to DEVNULL, to keep the token from being printed in (CI log) output.
        cmd = f"{cli_path} auth token".split()
        # pylint: disable-next=subprocess-run-check ; we want the return code here and don't want to raise when non-zero
        if bool(subprocess.run(cmd, stdout=subprocess.DEVNULL).returncode):
            raise SystemExit(" [!] A Phylum API key is required to continue.")

    def get_project_url(self, project_id: str) -> str:
        """Construct a project URL from a given project ID and return it.

        The project URL is used to access a particular project in the web UI, by label, and optionally with a group.
        """
        query = {"label": self.phylum_label}
        yaml = YAML()
        settings_dict = yaml.load(self.phylum_project_file.read_text(encoding="utf-8"))
        group_name = settings_dict.get("group_name")
        if group_name is not None:
            query.setdefault("group", group_name)
        query_params = urllib.parse.urlencode(query, safe="/", quote_via=urllib.parse.quote)
        project_url = f"https://app.phylum.io/projects/{project_id}?{query_params}"
        print(f" [+] Project URL: {project_url}")
        return project_url

    def analyze(self, analysis: dict) -> ReturnCode:
        """Analyze the results gathered from passing a lockfile to `phylum analyze`."""
        if self.args.all_deps:
            print(" [+] Considering all current dependencies ...")
            pkgs = analysis.get("packages", [])
            packages = [PackageDescriptor(pkg.get("name"), pkg.get("version"), pkg.get("type")) for pkg in pkgs]
            print(f" [+] {len(packages)} current dependencies")
            risk_data = self.parse_risk_data(analysis, packages)
        else:
            print(" [+] Only considering newly added dependencies ...")
            packages = self.get_new_deps()
            print(f" [+] {len(packages)} newly added dependencies")
            risk_data = self.parse_risk_data(analysis, packages)

        returncode = ReturnCode.SUCCESS
        output = ""
        if self.gbl_failed and not self.gbl_incomplete:
            print(" [!] The analysis is complete and there were failures")
            returncode = ReturnCode.FAILURE
            output = FAILED_COMMENT
            for line in risk_data:
                output += line

        if self.gbl_incomplete:
            incomplete_pkg_count = len(self.incomplete_pkgs)
            print(f" [+] {incomplete_pkg_count} packages were incomplete in the analysis results")
            if self.gbl_failed:
                print(" [!] There were failures in one or more completed packages")
                returncode = ReturnCode.FAILURE_INCOMPLETE
                output = INCOMPLETE_WITH_FAILURE_COMMENT_TEMPLATE.substitute(count=incomplete_pkg_count)
                for line in risk_data:
                    output += line
            else:
                print(" [+] There were no failures in the packages that have completed so far")
                returncode = ReturnCode.INCOMPLETE
                output = INCOMPLETE_COMMENT_TEMPLATE.substitute(count=incomplete_pkg_count)

        if not self.gbl_failed and not self.gbl_incomplete:
            print(" [+] The analysis is complete and there were NO failures")
            returncode = ReturnCode.SUCCESS
            output = SUCCESS_COMMENT

        project_id = analysis.get("project", "00000000-0000-0000-0000-000000000000")
        project_url = self.get_project_url(project_id)
        output += f"\n[View this project in the Phylum UI]({project_url})"
        self._analysis_output = output

        return returncode

    def parse_risk_data(self, analysis_results: dict, packages: Packages) -> List[str]:
        """Parse risk packages from a Phylum analysis.

        Packages that are in a completed analysis state will be included in the risk score report.
        Packages that have not completed analysis will be included with other incomplete packages
        and the overall PR will be allowed to pass, but with a note about re-running again later.
        """
        analysis_pkgs = analysis_results.get("packages", [])
        project_thresholds = analysis_results.get("thresholds", {})
        risk_scores = []
        for package in packages:
            for phylum_pkg in analysis_pkgs:
                if phylum_pkg.get("name") == package.name and phylum_pkg.get("version") == package.version:
                    if phylum_pkg.get("status") == "complete":
                        risk_score = self.check_risk_scores(phylum_pkg, project_thresholds)
                        if risk_score:
                            risk_scores.append(risk_score)
                    elif phylum_pkg.get("status") == "incomplete":
                        self.incomplete_pkgs.append(package)
                        self.gbl_incomplete = True

        return risk_scores

    def check_risk_scores(self, package_result: dict, project_thresholds: dict) -> Optional[str]:
        """Check risk scores of a package against project or user-provided thresholds.

        Thresholds provided via the corresponding option will take precedence over the project defined threshold value.
        If a package has a risk score below the threshold, set the fail flag and generate the markdown output.
        Return None when the package meets or exceeds all threshold values.
        """
        failed_flag = False
        risk_vectors = package_result.get("riskVectors", {})
        issue_flags = []

        fail_string = f"\n### Package: `{package_result.get('name')}@{package_result.get('version')}` failed.\n"
        fail_string += "|Risk Domain|Identified Score|Requirement|Requirement Source|\n"
        fail_string += "|-----------|----------------|-----------|------------------|\n"

        for threshold_option, risk_domain in PROJECT_THRESHOLD_OPTIONS.items():
            pti = self._get_project_threshold_info(project_thresholds, threshold_option)
            # The `RiskDomain` dataclass and this logic can be simplified once the API standardizes the use of names
            # so that risk domains are referenced with the same name everywhere (e.g., vulnerability/vulnerabilities
            # and malicious_code/malicious)
            potential_names = [risk_domain.package_name, risk_domain.project_name]
            for potential_name in potential_names:
                vul = risk_vectors.get(potential_name)
                if vul is not None:
                    break
            vul = vul if vul is not None else 1.0
            if vul < pti.threshold:
                failed_flag = True
                issue_flags.extend(potential_names)
                fail_string += f"|{risk_domain.output_name}|{vul*100:.0f}|{pti.threshold*100:.0f}|{pti.req_src}|\n"

        fail_string += "\n"
        fail_string += "#### Issues Summary\n"
        fail_string += "|Risk Domain|Risk Level|Title|\n"
        fail_string += "|-----------|----------|-----|\n"

        issue_list = build_issues_list(package_result, issue_flags)
        for domain, severity, title in issue_list:
            fail_string += f"|{domain}|{severity}|{title}|\n"

        if failed_flag:
            self.gbl_failed = True
            return fail_string
        return None

    def _get_project_threshold_info(self, project_thresholds: dict, threshold_type: str) -> ProjectThresholdInfo:
        """Determine the project threshold info in effect for a given threshold type.

        Thresholds for the five risk domains can be set individually in several ways.
        They can be set at the Phylum project level from either the Phylum CLI or the web UI.
        They can be set by `phylum-ci` options (e.g., `--vul-threshold`), which are distinguished by `threshold_type`.
        The default is to use the project level setting unless overridden by a value specified by a `phylum-ci` option.
        A default secure value will be used when neither of these sources are used to set the value.
        """
        threshold = getattr(self.args, threshold_type, None)
        req_src = "phylum-ci option"
        if threshold is None:
            risk_domain: RiskDomain = PROJECT_THRESHOLD_OPTIONS.get(threshold_type, RiskDomain("", "", ""))
            threshold = project_thresholds.get(risk_domain.project_name)
            req_src = "project per-axis threshold"
            if threshold is None:
                threshold = 1.0
                req_src = "N/A (fail safe)"

            total_threshold = project_thresholds.get("total", 0.0)
            if threshold < total_threshold:
                threshold = total_threshold
                req_src = "project total threshold"
        else:
            # The project risk threshold values returned by the analysis are normalized to [0.0, 1.0].
            # They are converted internally like this because it is more natural to ask users for input
            # as an integer in the range of [0, 100].
            threshold /= 100
        pti = ProjectThresholdInfo(threshold=threshold, req_src=req_src)
        return pti


# Type alias
CIEnvs = List[CIBase]


def build_issues_list(package_result: dict, issue_flags: List[str]) -> List[Tuple[str, str, str]]:
    """Build a list of issues from a given package's result object and return it."""
    issues = []
    pkg_issues = package_result.get("issues", [])
    for flag in issue_flags:
        for pkg_issue in pkg_issues:
            if flag == pkg_issue.get("domain"):
                domain = pkg_issue.get("domain")
                severity = pkg_issue.get("severity")
                title = pkg_issue.get("title")
                issues.append((domain, severity, title))
    return issues



"""Console script for phylum-ci."""
import argparse
import json
import os
import pathlib
import subprocess
import sys
from typing import List, Optional, Sequence, Tuple

from phylum import __version__
from phylum.ci import SCRIPT_NAME
from phylum.ci.ci_base import CIBase, CIEnvs
from phylum.ci.ci_github import CIGitHub
from phylum.ci.ci_gitlab import CIGitLab
from phylum.ci.ci_none import CINone
from phylum.ci.ci_precommit import CIPreCommit
from phylum.ci.common import ReturnCode
from phylum.constants import TOKEN_ENVVAR_NAME
from phylum.init.cli import get_target_triple, version_check


def detect_ci_platform(args: argparse.Namespace, remainder: List[str]) -> CIBase:
    """Detect CI platform via known CI-based environment variables."""
    ci_envs: CIEnvs = []

    # Detect GitLab CI
    if os.getenv("GITLAB_CI") == "true":
        print(" [+] CI environment detected: GitLab CI")
        ci_envs.append(CIGitLab(args))

    # Detect GitHub Actions
    if os.getenv("GITHUB_ACTIONS") == "true":
        print(" [+] CI environment detected: GitHub Actions")
        ci_envs.append(CIGitHub(args))

    # Detect Python pre-commit environment
    # This might be a naive strategy for detecting the `pre-commit` case, but there is at least
    # an attempt, via a pre-requisite check, to ensure all the extra arguments are staged files.
    if any(remainder):
        print(" [+] Extra arguments provided. Assuming a Python `pre-commit` working environment.")
        ci_envs.append(CIPreCommit(args, remainder))

    if len(ci_envs) > 1:
        ci_platform_names = ", ".join(ci_env.ci_platform_name for ci_env in ci_envs)
        raise SystemExit(f" [!] Multiple CI environments detected: {ci_platform_names}")
    if len(ci_envs) == 1:
        ci_env = ci_envs[0]
    else:
        print(" [+] No CI environment detected")
        ci_env = CINone(args)

    return ci_env


def get_phylum_analysis(ci_env: CIBase) -> dict:
    """Analyze a project lockfile from a given CI environment with the phylum CLI and return the analysis."""
    print(" [*] Performing analysis ...")
    cmd = f"{ci_env.cli_path} analyze -l {ci_env.phylum_label} --verbose --json {ci_env.lockfile}".split()
    try:
        analysis_result = subprocess.run(cmd, check=True, capture_output=True, text=True).stdout
    except subprocess.CalledProcessError as err:
        # The Phylum project can set the CLI to "fail the build" if threshold requirements are not met.
        # This causes the return code to be non-zero and lands us here. Check for this case to proceed.
        if "failed threshold requirements" in err.stderr:
            analysis_result = err.stdout
        else:
            print(f" [!] stderr: {err.stderr}")
            raise
    analysis = json.loads(analysis_result)
    return analysis


def threshold_check(threshold_in: str) -> int:
    """Check a given threshold for validity and return it as an int."""
    msg = "threshold must be an integer between 0 and 100, inclusive"
    try:
        threshold_out = int(threshold_in)
    except ValueError as err:
        raise argparse.ArgumentTypeError(msg) from err
    if threshold_out not in range(101):
        raise argparse.ArgumentTypeError(msg)
    return threshold_out


def get_args(args: Optional[Sequence[str]] = None) -> Tuple[argparse.Namespace, List[str]]:
    """Get the arguments from the command line and return them."""
    parser = argparse.ArgumentParser(
        prog=SCRIPT_NAME,
        description="Use Phylum to analyze dependencies in a CI environment",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"{SCRIPT_NAME} {__version__}",
    )

    analysis_group = parser.add_argument_group(title="Lockfile Analysis Options")
    analysis_group.add_argument(
        "-l",
        "--lockfile",
        type=pathlib.Path,
        help="""Path to the package lockfile to analyze. If not specified, an attempt will be made to automatically
            detect the lockfile. Some lockfile types (e.g., Python/pip `requirements.txt`) are ambiguous in that they
            can be named differently and may or may not contain strict dependencies. In these cases, it is best to
            specify an explicit lockfile path.""",
    )
    analysis_group.add_argument(
        "-a",
        "--all-deps",
        action="store_true",
        help="Specify this flag to consider all dependencies in analysis results instead of just the newly added ones.",
    )
    analysis_group.add_argument(
        "-f",
        "--force-analysis",
        action="store_true",
        help="Specify this flag to force analysis, even when the lockfile has not changed.",
    )
    analysis_group.add_argument(
        "-k",
        "--phylum-token",
        dest="token",
        help=f"""Phylum user token. Can also specify this option's value by setting the `{TOKEN_ENVVAR_NAME}`
            environment variable. The value specified with this option takes precedence when both are provided.
            Leave this option and it's related environment variable unspecified to either (1) use an existing token
            already set in the Phylum config file or (2) to manually populate the token with a `phylum auth login` or
            `phylum auth register` command after install.""",
    )

    threshold_group = parser.add_argument_group(
        title="Risk Domain Threshold Options",
        description="""Thresholds for the five risk domains may already be set at the Phylum project level.
            They can be set differently here for CI environments to "fail the build."
            The default is to use the project level setting unless overridden by a value specified in this group.
            Values must be an integer between 0 and 100, inclusive. Setting the value to zero (0) has the
            effect of disabling analysis against that risk domain. See "Phylum Risk Domains" documentation for more
            detail: https://docs.phylum.io/docs/phylum-package-score#risk-domains""",
    )
    threshold_group.add_argument(
        "-u",
        "--vul-threshold",
        type=threshold_check,
        help="v(u)lnerability risk score threshold value.",
    )
    threshold_group.add_argument(
        "-m",
        "--mal-threshold",
        type=threshold_check,
        help="(m)alicious Code risk score threshold value.",
    )
    threshold_group.add_argument(
        "-e",
        "--eng-threshold",
        type=threshold_check,
        help="(e)ngineering risk score threshold value.",
    )
    threshold_group.add_argument(
        "-c",
        "--lic-threshold",
        type=threshold_check,
        help="li(c)ense risk score threshold value.",
    )
    threshold_group.add_argument(
        "-o",
        "--aut-threshold",
        type=threshold_check,
        help="auth(o)r risk score threshold value.",
    )

    cli_group = parser.add_argument_group(
        title="Phylum CLI Options",
        description="""Use the options here to control the Phylum CLI version in use during analysis.
            Examples of when this may be useful are: for troubleshooting, maintaining a consistent evironment,
            ensuring the latest version, or reverting to a previous version when the installed one causes an issue.""",
    )
    cli_group.add_argument(
        "-r",
        "--phylum-release",
        dest="version",
        default="latest",
        type=version_check,
        help="""The version of the Phylum CLI to install, when one is not already installed. Can be specified as
            `latest` or a specific tagged release, with or without the leading `v`.""",
    )
    cli_group.add_argument(
        "-t",
        "--target",
        default=get_target_triple(),
        help="The target platform type where the CLI will be installed.",
    )
    cli_group.add_argument(
        "-i",
        "--force-install",
        action="store_true",
        help="""Specify this flag to ensure the specified Phylum CLI release version is the one that is installed.
            Otherwise, any existing version will be used.""",
    )

    return parser.parse_known_args(args=args)


def main(args: Optional[Sequence[str]] = None) -> int:
    """Main entrypoint."""
    parsed_args, remainder_args = get_args(args=args)

    # Detect which CI environment, if any, we are in
    ci_env = detect_ci_platform(parsed_args, remainder_args)

    # Bail early if there are no changes to the lockfile
    print(f" [+] lockfile in use: {ci_env.lockfile}")
    if parsed_args.force_analysis:
        print(" [+] `--force-analysis` flag specified. Proceeding with analysis ...")
    else:
        if ci_env.is_lockfile_changed:
            print(" [+] The lockfile has changed. Proceeding with analysis ...")
        else:
            print(" [+] The lockfile has not changed. Nothing to do.")
            return 0

    # Generate a label to use for analysis and report it
    print(f" [+] phylum_label: {ci_env.phylum_label}")

    # Check for the existence of the CLI and install it if needed
    ci_env.init_cli()

    # Analyze current project lockfile with phylum CLI
    analysis = get_phylum_analysis(ci_env)

    # Review analysis results to determine the overall state
    return_code = ci_env.analyze(analysis)
    print(f" [-] Return code: {return_code}")

    # Output the results of the analysis
    ci_env.post_output()

    # Don't return a failure code if the results are unknown at this point
    if return_code == ReturnCode.INCOMPLETE:
        return 0
    return return_code.value


def script_main() -> None:
    """Script entry point.

    The only point of this function is to ensure the proper exit code is set when called from the script entry point."""
    sys.exit(main())


if __name__ == "__main__":
    script_main()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['phylum',
 'phylum.ci',
 'phylum.init',
 'tests',
 'tests.functional',
 'tests.unit']

package_data = \
{'': ['*']}

install_requires = \
['connect-markdown-renderer',
 'cryptography',
 'packaging',
 'requests',
 'ruamel.yaml']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata']}

entry_points = \
{'console_scripts': ['phylum-ci = phylum.ci.cli:script_main',
                     'phylum-init = phylum.init.cli:main']}

setup_kwargs = {
    'name': 'phylum',
    'version': '0.13.2',
    'description': 'Utilities for integrating Phylum into CI pipelines (and beyond)',
    'long_description': '# phylum-ci\n[![PyPI](https://img.shields.io/pypi/v/phylum)](https://pypi.org/project/phylum/)\n![PyPI - Status](https://img.shields.io/pypi/status/phylum)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/phylum)](https://pypi.org/project/phylum/)\n[![GitHub](https://img.shields.io/github/license/phylum-dev/phylum-ci)][license]\n[![GitHub issues](https://img.shields.io/github/issues/phylum-dev/phylum-ci)][issues]\n![GitHub last commit](https://img.shields.io/github/last-commit/phylum-dev/phylum-ci)\n[![GitHub Workflow Status (branch)][workflow_shield]][workflow_test]\n[![Contributor Covenant](https://img.shields.io/badge/Contributor%20Covenant-2.1-4baaaa.svg)][CoC]\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit)][pre-commit]\n\nUtilities for integrating Phylum into CI pipelines (and beyond)\n\n[license]: https://github.com/phylum-dev/phylum-ci/blob/main/LICENSE\n[issues]: https://github.com/phylum-dev/phylum-ci/issues\n[workflow_shield]: https://img.shields.io/github/workflow/status/phylum-dev/phylum-ci/Test/main?label=Test&logo=GitHub\n[workflow_test]: https://github.com/phylum-dev/phylum-ci/actions/workflows/test.yml\n[CoC]: https://github.com/phylum-dev/phylum-ci/blob/main/CODE_OF_CONDUCT.md\n[pre-commit]: https://github.com/pre-commit/pre-commit\n[contributing]: https://github.com/phylum-dev/phylum-ci/blob/main/CONTRIBUTING.md\n[changelog]: https://github.com/phylum-dev/phylum-ci/blob/main/CHANGELOG.md\n[security]: https://github.com/phylum-dev/phylum-ci/blob/main/docs/security.md\n\n## Installation and usage\n\n### Installation\n\nThe `phylum` Python package is pip installable for the environment of your choice:\n\n```sh\npip install phylum\n```\n\nIt can also be installed in an isolated environment with the excellent [`pipx` tool](https://pypa.github.io/pipx/):\n\n```sh\n# Globally install the app(s) on your system in an isolated virtual environment for the package\npipx install phylum\n\n# Use the apps from the package in an ephemeral environment\npipx run --spec phylum phylum-init <options>\npipx run --spec phylum phylum-ci <options>\n```\n\nThese installation methods require Python 3.7+ to run. For a self contained environment, consider using the Docker\nimage as described below.\n\n### Usage\n\nThe `phylum` Python package exposes its functionality with a command line interface (CLI).\nTo view the options available from the CLI, print the help message from one of the scripts provided as entry points:\n\n```sh\nphylum-init -h\nphylum-ci -h\n```\n\nThe functionality can also be accessed by calling the module:\n\n```sh\npython -m phylum.init -h\npython -m phylum.ci -h\n```\n\nThe functionality is also exposed in the form of a Docker image:\n\n```sh\n# Get the `latest` tagged image\ndocker pull phylumio/phylum-ci\n\n# View the help\ndocker run --rm phylumio/phylum-ci phylum-ci --help\n\n# Export a Phylum token (e.g., from `phylum auth token`)\nexport PHYLUM_API_KEY=$(phylum auth token)\n\n# Run it from a git repo directory containing a `.phylum_project` and a lockfile\ndocker run -it --rm -e PHYLUM_API_KEY --mount type=bind,src=$(pwd),dst=/phylum -w /phylum phylumio/phylum-ci\n```\n\nThe Docker image contains `git` and the installed `phylum` Python package.\nIt also contains an installed version of the Phylum CLI.\nAn advantage of using the Docker image is that the complete environment is packaged and made available with components\nthat are known to work together.\n\nWhen using the `latest` tagged image, the version of the Phylum CLI is the `latest` available.\nThere are additional image tag options available to specify a specific release of the `phylum-ci` project and a specific\nversion of the Phylum CLI, in the form of `<phylum-ci version>-CLIv<Phylum CLI version>`. Here are image tag examples:\n\n```sh\n# Get the most current release of *both* `phylum-ci` and the Phylum CLI\ndocker pull phylumio/phylum-ci:latest\n\n# Get the image with `phylum-ci` version 0.13.0 and Phylum CLI version 3.8.0\ndocker pull phylumio/phylum-ci:0.13.0-CLIv3.8.0\n```\n\n#### `phylum-init` Script Entry Point\n\nThe `phylum-init` script can be used to fetch and install the Phylum CLI.\nIt will attempt to install the latest released version of the CLI but can be specified to fetch a specific version.\nIt will attempt to automatically determine the correct CLI release, based on the platform where the script is run, but\na specific release target can be specified.\nIt will accept a Phylum token from an environment variable or specified as an option, but will also function in the case\nthat no token is provided. This can be because there is already a token set that should continue to be used or because\nno token exists and one will need to be manually created or set, after the CLI is installed.\n\nThe options for `phylum-init`, automatically updated to be current for the latest release:\n\n![phylum-init options](https://raw.githubusercontent.com/phylum-dev/phylum-ci/main/docs/img/phylum-init_options.svg)\n\n#### `phylum-ci` Script Entry Point\n\nThe `phylum-ci` script is for analyzing lockfile changes.\nThe script can be used locally or from within a Continuous Integration (CI) environment.\nIt will attempt to detect the CI platform based on the environment from which it is run and act accordingly.\nThe current CI platforms/environments supported are:\n\n* GitLab CI\n  * See the [GitLab CI Integration documentation][gitlab_docs] for more info\n\n* GitHub Actions\n  * See the [GitHub Actions Integration documentation][github_docs] for more info\n\n* Git `pre-commit` Hooks\n  * See the [Git `pre-commit` Integration documentation][precommit_docs] for more info\n\n* None (local use)\n  * This is the "fall-through" case used when no other environment is detected\n  * Can be useful to analyze lockfiles locally, prior to or after submitting a pull/merge request (PR/MR) to a CI system\n    * Establishing a successful submission prior to submitting a PR/MR to a CI system\n    * Troubleshooting after submitting a PR/MR to a CI system and getting unexpected results\n\nThe options for `phylum-ci`, automatically updated to be current for the latest release:\n\n![phylum-ci options](https://raw.githubusercontent.com/phylum-dev/phylum-ci/main/docs/img/phylum-ci_options.svg)\n\n[gitlab_docs]: https://github.com/phylum-dev/phylum-ci/blob/main/docs/sync/gitlab_ci.md\n[github_docs]: https://github.com/phylum-dev/phylum-ci/blob/main/docs/sync/github_actions.md\n[precommit_docs]: https://github.com/phylum-dev/phylum-ci/blob/main/docs/sync/git_precommit.md\n\n## License\n\nMIT - with complete text available in the [LICENSE][license] file.\n\n## Contributing\n\nSuggestions and help are welcome. Feel free to open an issue or otherwise contribute.\nMore information is available on the [contributing documentation][contributing] page.\n\n## Code of Conduct\n\nEveryone participating in the `phylum-ci` project, and in particular in the issue tracker and pull requests, is\nexpected to treat other people with respect and more generally to follow the guidelines articulated in the\n[Code of Conduct][CoC].\n\n## Security Disclosures\n\nFound a security issue in this repository? See the [security policy][security]\nfor details on coordinated disclosure.\n\n## Change log\n\nAll notable changes to this project are documented in the [CHANGELOG][changelog].\n\nThe format of the change log is based on [Keep a Changelog](https://keepachangelog.com/en/1.0.0/),\nand this project adheres to [Semantic Versioning](https://semver.org/spec/v2.0.0.html).\nThe entries in the changelog are primarily automatically generated through the use of\n[conventional commits](https://www.conventionalcommits.org) and the\n[Python Semantic Release](https://python-semantic-release.readthedocs.io/en/latest/index.html) tool.\nHowever, some entries may be manually edited, where it helps for clarity and understanding.\n',
    'author': 'Phylum, Inc.',
    'author_email': 'engineering@phylum.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://phylum.io/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Top-level package for phylum."""
# TODO: Use only the standard library form (importlib.metadata) after Python 3.7 support is dropped
#       https://github.com/phylum-dev/phylum-ci/issues/18
try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata


PKG_METADATA = importlib_metadata.metadata(__name__)

__version__ = importlib_metadata.version(__name__)
__author__ = PKG_METADATA["Author"]
__email__ = PKG_METADATA["Author-email"]

PKG_NAME = PKG_METADATA["Name"]
PKG_SUMMARY = PKG_METADATA["Summary"]



"""Test the package metadata."""

import sys

from packaging.version import Version
from phylum import PKG_NAME, PKG_SUMMARY, __author__, __email__, __version__

from ..constants import PYPROJECT


def test_project_version():
    """Ensure the source version can be normalized as specified in PEP-440."""
    package_version = Version(__version__)
    source_version = Version(PYPROJECT.get("tool", {}).get("poetry", {}).get("version"))
    assert package_version == source_version


def test_python_version():
    """Ensure the python version used to test is a supported version."""
    supported_minor_versions = (7, 8, 9, 10)
    python_version = sys.version_info
    assert python_version.major == 3, "Only Python 3 is supported"
    assert python_version.minor in supported_minor_versions, "Attempting to run unsupported Python version"


def test_author_email_metadata():
    """Ensure the project and package metadata for author and email match and are correct."""
    assert __author__ == "Phylum, Inc.", "The company name should be used instead of individual developers"
    assert __email__ == "engineering@phylum.io", "The engineering group email account should be used"
    # Package authors in Poetry are specified as a list of "name <email>" entries
    expected_poetry_author = f"{__author__} <{__email__}>"
    poetry_authors = PYPROJECT.get("tool", {}).get("poetry", {}).get("authors", [])
    assert expected_poetry_author in poetry_authors, "Package author/email should be defined in pyproject.toml only"
    assert len(poetry_authors) == 1, "There should only be one author - the company, with it's engineering group email"


def test_package_name():
    """Ensure the package name is traced through from the pyproject.toml definition."""
    expected_pkg_name = PYPROJECT.get("tool", {}).get("poetry", {}).get("name", "")
    assert expected_pkg_name == PKG_NAME, "The package name should be defined in pyproject.toml only"


def test_package_description():
    """Ensure the package description is traced through from the pyproject definition."""
    expected_pkg_desc = PYPROJECT.get("tool", {}).get("poetry", {}).get("description", "")
    assert expected_pkg_desc == PKG_SUMMARY, "The package description should be defined in pyproject.toml only"



""""Test the phylum-init command line interface (CLI)."""
import subprocess
import sys

from phylum import __version__
from phylum.init import SCRIPT_NAME

from ..constants import PYPROJECT


def test_run_as_module():
    """Ensure the CLI can be called as a module.

    This is the `python -m <module_name>` format to "run library module as a script."
    NOTE: The <module_name> is specified as the dotted path to the package where the `__main__.py` module exists.
    """
    cmd = f"{sys.executable} -m phylum.init --help".split()
    ret = subprocess.run(cmd)
    assert ret.returncode == 0, "Running the package as a module failed"


def test_run_as_script():
    """Ensure the CLI can be called by it's script entry point."""
    scripts = PYPROJECT.get("tool", {}).get("poetry", {}).get("scripts", {})
    assert scripts, "There should be at least one script entry point"
    assert SCRIPT_NAME in scripts, f"The {SCRIPT_NAME} script should be a defined entry point"
    ret = subprocess.run([SCRIPT_NAME, "-h"])
    assert ret.returncode == 0, f"{SCRIPT_NAME} entry point failed"


def test_version_option():
    """Ensure the correct program name and version is displayed when using the `--version` option."""
    # The argparse module adds a newline to the output
    expected_output = f"{SCRIPT_NAME} {__version__}\n"
    cmd = f"{sys.executable} -m phylum.init --version".split()
    ret = subprocess.run(cmd, capture_output=True, encoding="utf-8")
    assert not ret.stderr, "Nothing should be written to stderr"
    assert ret.returncode == 0, "A non-successful return code was provided"
    assert ret.stdout == expected_output, "Output did not match expected input"



"""Provide constants for use throughout the package."""
import string
import textwrap

from phylum.ci.common import RiskDomain

# The common Phylum header that must exist as the first text in the first line of all analysis output
PHYLUM_HEADER = "# Phylum OSS Supply Chain Risk Analysis"

# NOTE: All multi-line strings are indented by two levels on purpose, to ensure they line up correctly when used with
#       each other in templates and are all fully left justified after applying `textwrap.dedent` for normalization.

SUCCESS_DETAILS = "The Phylum risk analysis is complete and did not identify any issues."

# Expandable HTML providing information on why there was a failure
FAILURE_DETAILS = """
        <details>
        <summary>Background</summary>
        <br />
        This repository analyzes the risk of new dependencies. An administrator of
        this repository has set score requirements for Phylum's five risk domains.
        <br /><br />
        If you see this comment, one or more dependencies added to the
        package manager lockfile have failed Phylum's risk analysis.
        </details>
        """.strip()

# String template providing information about an incomplete analysis.
# Substitution variable(s) should be accounted for when using string variables containing this text.
INCOMPLETE_DETAILS = """
        The analysis contains $count package(s) Phylum has not yet processed,
        preventing a complete risk analysis. Phylum is processing these
        packages currently and should complete soon.
        Please wait for up to 30 minutes, then re-run the analysis.
        """

# Headers for distinct comment types:
FAILED_COMMENT = textwrap.dedent(
    f"""
        {PHYLUM_HEADER} - FAILED

        {FAILURE_DETAILS}
        """
)

SUCCESS_COMMENT = textwrap.dedent(
    f"""
        {PHYLUM_HEADER} - SUCCESS

        {SUCCESS_DETAILS}
        """
)

# Substitution variable(s) should be accounted for when using this template
INCOMPLETE_WITH_FAILURE_COMMENT_TEMPLATE = string.Template(
    textwrap.dedent(
        f"""
        {PHYLUM_HEADER} - INCOMPLETE WITH FAILURE
        {INCOMPLETE_DETAILS}
        {FAILURE_DETAILS}
        """
    )
)

# Substitution variable(s) should be accounted for when using this template
INCOMPLETE_COMMENT_TEMPLATE = string.Template(
    textwrap.dedent(
        f"""
        {PHYLUM_HEADER} - INCOMPLETE
        {INCOMPLETE_DETAILS}
        """
    )
)

# These are the project threshold options.
# Keys are the risk domain threshold options, as provided/known by argparse.
# Values are a RiskDomain dataclass object containing:
#   * the descriptive name
#   * key name returned from a Phylum analysis as known by the overall project threshold mapping
#   * key name returned from a Phylum analysis as known by an individual package riskVectors mapping
PROJECT_THRESHOLD_OPTIONS = {
    "vul_threshold": RiskDomain("Software Vulnerability", "vulnerability", "vulnerabilities"),
    "mal_threshold": RiskDomain("Malicious Code", "malicious", "malicious_code"),
    "eng_threshold": RiskDomain("Engineering", "engineering", "engineering"),
    "lic_threshold": RiskDomain("License", "license", "license"),
    "aut_threshold": RiskDomain("Author", "author", "author"),
}
